//
//  split_line.cpp
//  STL string tokenizer
//
//  Created by Clemens Wacha.
//  Version 1.0
//  Copyright (c) 2005 Clemens Wacha. All rights reserved.
//

#include <iostream>
#include <string>
#include <vector>


using namespace std;

enum {
	SL_NORMAL,
	SL_ESCAPE,
	SL_SAFEMODE,
	SL_SAFEESCAPE,
	SL_COMMENT,
} SPLIT_LINE_STATE;
	
// splits line into tokens and stores them in ret. Supports delimiters, escape characters, ignores special characters between safemode_char and between comment_char and linend '\n'.
// returns SPLIT_LINE_STATE the parser was in when returning
int split_line(std::vector<std::string>& ret, std::string& line, const std::string& delimiters = " \t\r\n", char escape_char = '\\', char safemode_char = '"', char comment_char = '\0', int start_state = SL_NORMAL);

int split_line(vector<string>& ret, string& line, const string& delimiters, char escape_char, char safemode_char, char comment_char, int start_state) {
	unsigned int state = start_state;
	unsigned int i = 0;
	string token;
	
	if(start_state != SL_NORMAL && ret.size() > 0) {
		token = ret[ret.size()-1];
		ret.pop_back();
	}
	
	while(i < line.size()) {
		switch(state) {
			case SL_NORMAL:
				if(line[i] == escape_char) {
					state = SL_ESCAPE;
				} else if(line[i] == safemode_char) {
					state = SL_SAFEMODE;
				} else if(line[i] == comment_char) {
					/// FINISH
					if(token.size() > 0) {
						ret.push_back(token);
						token.clear();
					}
					token += line[i];	// EAT
					state = SL_COMMENT;
				} else if(delimiters.find(line[i]) != string::npos) {
					// line[i] is a delimiter
					/// FINISH
					if(token.size() > 0) {
						ret.push_back(token);
						token.clear();
					}
				} else {
					token += line[i];	// EAT
				}
				break;
			case SL_ESCAPE:
				if(line[i] == 'n') token += '\n';
				else if(line[i] == 't') token += '\t';
				else if(line[i] == 'v') token += '\v';
				else if(line[i] == 'b') token += '\b';
				else if(line[i] == 'r') token += '\r';
				else if(line[i] == 'f') token += '\f';
				else if(line[i] == 'a') token += '\a';
				else if(line[i] == '?') token += '\?';
				else token += line[i];	// EAT
				state = SL_NORMAL;
				break;
			case SL_SAFEMODE:
				if(line[i] == safemode_char) {
					state = SL_NORMAL;
				} else if(line[i] == escape_char) {
					state = SL_SAFEESCAPE;
				} else {
					token += line[i];	// EAT
				}
				break;
			case SL_SAFEESCAPE:
				if(line[i] == 'n') token += '\n';
				else if(line[i] == 't') token += '\t';
				else if(line[i] == 'v') token += '\v';
				else if(line[i] == 'b') token += '\b';
				else if(line[i] == 'r') token += '\r';
				else if(line[i] == 'f') token += '\f';
				else if(line[i] == 'a') token += '\a';
				else if(line[i] == '?') token += '\?';
				else token += line[i];	// EAT
				state = SL_SAFEMODE;
				break;
			case SL_COMMENT:
				if(line[i] == '\n') {
					/// FINISH
					if(token.size() > 0) {
						ret.push_back(token);
						token.clear();
					}
					state = SL_NORMAL;
				} else {
					token += line[i];	// EAT
				}
				break;
			default:
				// nothing
				break;
		}
		
		i++;	
	}

	/// FINISH
	if(token.size() > 0) {
		ret.push_back(token);
		token.clear();
	}
	return(state);
}


int main(int argc, char *argv[]) {
	vector<string> tokens;
	string line = "Writing    programs     \"in C++\"  	is   \
	 Fun!!";
	
	split_line(tokens, line);
	
	cout << "Tokens:" << endl;
	for(unsigned int i = 0; i < tokens.size(); i++) cout << "'" << tokens[i] << "'" << endl;
	

    system("PAUSE");
    return 0;
}
